@@ -76,6 +76,9 @@ pub enum CipherSuite {
     RSA2k,
     /// 4096 bit RSA with SHA512 and AES256
     RSA4k,
+
+    // If you add a variant here, be sure to update
+    // CipherSuite::variants below.
 }
 assert_send_and_sync!(CipherSuite);
 
@@ -86,6 +89,14 @@ impl Default for CipherSuite {
 }
 
 impl CipherSuite {
+    /// Returns an iterator over `CipherSuite`'s  variants.
+    pub fn variants() -> impl Iterator<Item=CipherSuite> {
+        use CipherSuite::*;
+
+        [ Cv25519, RSA3k, P256, P384, P521, RSA2k, RSA4k ]
+            .into_iter()
+    }
+
     /// Returns whether the currently selected cryptographic backend
     /// supports the encryption and signing algorithms that the cipher
     /// suite selects.
@@ -1715,7 +1726,8 @@ mod tests {
             .set_primary_key_flags(KeyFlags::empty())
             .add_subkey(KeyFlags::empty().set_certification(), None, None)
             .generate().unwrap();
-        let sig_pkts = cert1.subkeys().next().unwrap().bundle().self_signatures[0].hashed_area();
+        let sig_pkts = cert1.subkeys().next().unwrap().bundle()
+            .self_signatures2().next().unwrap().hashed_area();
 
         match sig_pkts.subpacket(SubpacketTag::KeyFlags).unwrap().value() {
             SubpacketValue::KeyFlags(ref ks) => assert!(ks.for_certification()),
@@ -1793,7 +1805,7 @@ mod tests {
             .generate().unwrap();
 
         let key = cert.primary_key().key();
-        let sig = &cert.primary_key().bundle().self_signatures()[0];
+        let sig = &cert.primary_key().bundle().self_signatures2().next().unwrap();
         assert!(sig.key_alive(key, now).is_ok());
         assert!(sig.key_alive(key, now + 590 * s).is_ok());
         assert!(! sig.key_alive(key, now + 610 * s).is_ok());
