@@ -79,9 +79,12 @@
 //! [complicated semantics]: https://tools.ietf.org/html/rfc4880#section-5.2.3.3
 
 use std::time;
-use std::ops::Deref;
+use std::cmp::{self, Ordering};
+use std::ops::{Deref, DerefMut};
+use std::sync::Arc;
 
 use crate::{
+    cert::lazysigs::{LazySignatures, SigState},
     Error,
     packet::Signature,
     packet::Key,
@@ -110,25 +113,29 @@ use super::{
 /// description.
 #[derive(Debug, Clone, PartialEq)]
 pub struct ComponentBundle<C> {
-    pub(crate) component: C,
+    component: C,
 
-    pub(crate) hash_algo_security: HashAlgoSecurity,
+    pub(super) hash_algo_security: HashAlgoSecurity,
 
     // Self signatures.
-    pub(crate) self_signatures: Vec<Signature>,
+    pub(super) self_signatures: LazySignatures,
+
+    /// If set, is equal to `component`, and provides context to
+    /// verify primary key binding signatures.
+    backsig_signer: Option<Key<key::PublicParts, key::SubordinateRole>>,
 
     // Third-party certifications.  (In general, this will only be by
     // designated revokers.)
-    pub(crate) certifications: Vec<Signature>,
+    pub(super) certifications: Vec<Signature>,
 
     // Attestation key signatures.
-    pub(crate) attestations: Vec<Signature>,
+    pub(super) attestations: LazySignatures,
 
     // Self revocations.
-    pub(crate) self_revocations: Vec<Signature>,
+    pub(super) self_revocations: LazySignatures,
 
     // Third-party revocations (e.g., designated revokers).
-    pub(crate) other_revocations: Vec<Signature>,
+    pub(super) other_revocations: Vec<Signature>,
 }
 assert_send_and_sync!(ComponentBundle<C> where C);
 
@@ -178,6 +185,29 @@ impl<C> Deref for ComponentBundle<C>
 }
 
 impl<C> ComponentBundle<C> {
+    /// Creates a new component.
+    ///
+    /// Should only be used from the cert parser.  However, we cannot
+    /// use `pub(in ...)` because the cert parser isn't an ancestor of
+    /// this module.
+    pub(crate) fn new(component: C,
+                      hash_algo_security: HashAlgoSecurity,
+                      sigs: Vec<Signature>,
+                      primary_key: Arc<Key<key::PublicParts, key::PrimaryRole>>)
+           -> ComponentBundle<C>
+    {
+        ComponentBundle {
+            component,
+            hash_algo_security,
+            self_signatures: LazySignatures::new(primary_key.clone()),
+            backsig_signer: None,
+            certifications: sigs,
+            attestations: LazySignatures::new(primary_key.clone()),
+            self_revocations: LazySignatures::new(primary_key),
+            other_revocations: vec![],
+        }
+    }
+
     /// Returns a reference to the bundle's component.
     ///
     /// # Examples
@@ -245,7 +275,9 @@ impl<C> ComponentBundle<C> {
       /// is unfortunately monomorphized for every `C`.  Prevent
       /// this by moving the code to a function independent of `C`.
       fn find_binding_signature<'s>(policy: &dyn Policy,
-                                    self_signatures: &'s [Signature],
+                                    self_signatures: &'s LazySignatures,
+                                    backsig_signer:
+                                    Option<&Key<key::PublicParts, key::SubordinateRole>>,
                                     hash_algo_security: HashAlgoSecurity,
                                     t: time::SystemTime)
                                     -> Result<&'s Signature>
@@ -256,15 +288,26 @@ impl<C> ComponentBundle<C> {
         // We want the newest signature that is older than `t`, or
         // that has been created at `t`.  So, search for `t`.
 
+        // We search all signatures without triggering the signature
+        // verification.  Later, we will verify the candidates, and
+        // reject bad signatures.
+        let unverified_self_signatures = self_signatures.as_slice_unverified();
+
         let i =
             // Usually, the first signature is what we are looking for.
             // Short circuit the binary search.
-            if Some(t) >= self_signatures.get(0)
-                              .and_then(|s| s.signature_creation_time())
+              if unverified_self_signatures.get(0)
+              .filter(|s| s.signature_creation_time().map(|c| t >= c)
+                      .unwrap_or(false))
+              .filter(
+                  // Verify the signature now.
+                  |_| matches!(self_signatures.verify_sig(0, backsig_signer),
+                               Ok(SigState::Good)))
+              .is_some()
             {
                 0
             } else {
-                match self_signatures.binary_search_by(
+                match unverified_self_signatures.binary_search_by(
                     |s| canonical_signature_order(
                         s.signature_creation_time(), Some(t)))
                 {
@@ -278,7 +321,7 @@ impl<C> ComponentBundle<C> {
                     // return index 1, 2, 3 or 4.
                     Ok(mut i) => {
                         while i > 0
-                            && self_signatures[i - 1].signature_creation_time()
+                            && unverified_self_signatures[i - 1].signature_creation_time()
                             == Some(t)
                         {
                             i -= 1;
@@ -307,7 +350,9 @@ impl<C> ComponentBundle<C> {
         // `t`.
         let mut error = None;
 
-        'next_sig: for s in self_signatures[i..].iter() {
+        'next_sig: for (j, s) in unverified_self_signatures[i..].iter()
+            .enumerate()
+        {
             if let Err(e) = s.signature_alive(t, time::Duration::new(0, 0)) {
                 // We know that t >= signature's creation time.  So,
                 // it is expired.  But an older signature might not
@@ -326,6 +371,13 @@ impl<C> ComponentBundle<C> {
                 continue;
             }
 
+            // Verify the signature now.
+            if ! matches!(self_signatures.verify_sig(i + j, backsig_signer),
+                          Ok(SigState::Good)) {
+                // Reject bad signatures.
+                continue;
+            }
+
             // The signature is good, but we may still need to verify the
             // back sig.
             if s.typ() == crate::types::SignatureType::SubkeyBinding &&
@@ -389,8 +441,11 @@ impl<C> ComponentBundle<C> {
         }
       }
 
-        find_binding_signature(policy, &self.self_signatures,
-                               self.hash_algo_security, t)
+        find_binding_signature(
+            policy,
+            &self.self_signatures,
+            self.backsig_signer.as_ref(),
+            self.hash_algo_security, t)
     }
 
     /// Returns the component's self-signatures.
@@ -414,12 +469,20 @@ impl<C> ComponentBundle<C> {
     /// for (i, ka) in cert.keys().enumerate() {
     ///     eprintln!("Key #{} ({}) has {:?} self signatures",
     ///               i, ka.fingerprint(),
-    ///               ka.bundle().self_signatures().len());
+    ///               ka.bundle().self_signatures2().count());
     /// }
     /// # Ok(()) }
     /// ```
+    // XXXv2: Rename this function.
+    pub fn self_signatures2(&self)
+                            -> impl Iterator<Item=&Signature> + Send + Sync {
+        self.self_signatures.iter_verified(self.backsig_signer.as_ref())
+    }
+
+    /// Returns the component's self-signatures.
+    #[deprecated(note = "Use self_signatures2 instead.")]
     pub fn self_signatures(&self) -> &[Signature] {
-        &self.self_signatures
+        &self.self_signatures.slice_verified(self.backsig_signer.as_ref())
     }
 
     /// Returns the component's third-party certifications.
@@ -443,10 +506,18 @@ impl<C> ComponentBundle<C> {
     /// for ua in cert.userids() {
     ///     eprintln!("User ID {} has {:?} unverified, third-party certifications",
     ///               String::from_utf8_lossy(ua.userid().value()),
-    ///               ua.bundle().certifications().len());
+    ///               ua.bundle().certifications2().count());
     /// }
     /// # Ok(()) }
     /// ```
+    // XXXv2: Rename this function.
+    pub fn certifications2(&self)
+                           -> impl Iterator<Item=&Signature> + Send + Sync {
+        self.certifications.iter()
+    }
+
+    /// Returns the component's third-party certifications.
+    #[deprecated(note = "Use certifications2 instead.")]
     pub fn certifications(&self) -> &[Signature] {
         &self.certifications
     }
@@ -473,12 +544,21 @@ impl<C> ComponentBundle<C> {
     /// for u in cert.userids() {
     ///     eprintln!("User ID {} has {:?} revocation certificates.",
     ///               String::from_utf8_lossy(u.userid().value()),
-    ///               u.bundle().self_revocations().len());
+    ///               u.bundle().self_revocations2().count());
     /// }
     /// # Ok(()) }
     /// ```
+    // XXXv2: Rename this function.
+    pub fn self_revocations2(&self)
+                             -> impl Iterator<Item=&Signature> + Send + Sync {
+        self.self_revocations.iter_verified(self.backsig_signer.as_ref())
+    }
+
+    /// Returns the component's revocations that were issued by the
+    /// certificate holder.
+    #[deprecated(note = "Use self_revocations2 instead.")]
     pub fn self_revocations(&self) -> &[Signature] {
-        &self.self_revocations
+        &self.self_revocations.slice_verified(self.backsig_signer.as_ref())
     }
 
     /// Returns the component's revocations that were issued by other
@@ -503,10 +583,19 @@ impl<C> ComponentBundle<C> {
     /// for u in cert.userids() {
     ///     eprintln!("User ID {} has {:?} unverified, third-party revocation certificates.",
     ///               String::from_utf8_lossy(u.userid().value()),
-    ///               u.bundle().other_revocations().len());
+    ///               u.bundle().other_revocations2().count());
     /// }
     /// # Ok(()) }
     /// ```
+    // XXXv2: Rename this function.
+    pub fn other_revocations2(&self)
+                              -> impl Iterator<Item=&Signature> + Send + Sync {
+        self.other_revocations.iter()
+    }
+
+    /// Returns the component's revocations that were issued by other
+    /// certificates.
+    #[deprecated(note = "Use other_revocations2 instead.")]
     pub fn other_revocations(&self) -> &[Signature] {
         &self.other_revocations
     }
@@ -547,7 +636,7 @@ impl<C> ComponentBundle<C> {
     pub fn attestations(&self)
                       -> impl Iterator<Item = &Signature> + Send + Sync
     {
-        self.attestations.iter()
+        self.attestations.iter_verified(None)
     }
 
     /// Returns all of the component's signatures.
@@ -583,11 +672,11 @@ impl<C> ComponentBundle<C> {
     pub fn signatures(&self)
                       -> impl Iterator<Item = &Signature> + Send + Sync
     {
-        self.self_revocations.iter()
-            .chain(self.self_signatures.iter())
-            .chain(self.attestations.iter())
-            .chain(self.certifications.iter())
-            .chain(self.other_revocations.iter())
+        self.self_revocations2()
+            .chain(self.self_signatures2())
+            .chain(self.attestations())
+            .chain(self.certifications2())
+            .chain(self.other_revocations2())
     }
 
     /// Returns the component's revocation status at time `t`.
@@ -626,10 +715,10 @@ impl<C> ComponentBundle<C> {
                 selfsig.signature_alive(t, time::Duration::new(0, 0)).is_ok());
         }
 
-        let check = |revs: &'a [Signature], sec: HashAlgoSecurity|
+        let check = |revs: &mut dyn Iterator<Item=&'a Signature>, sec: HashAlgoSecurity|
             -> Option<Vec<&'a Signature>>
         {
-            let revs = revs.iter().filter(|rev| {
+            let revs = revs.filter(|rev| {
                 if let Err(err) = policy.signature(rev, sec) {
                     t!("  revocation rejected by caller policy: {}", err);
                     false
@@ -686,14 +775,18 @@ impl<C> ComponentBundle<C> {
         };
 
         if let Some(revs)
-            = check(&self.self_revocations, self.hash_algo_security)
+            = check(&mut self.self_revocations.iter_verified(self.backsig_signer.as_ref()),
+                    self.hash_algo_security)
         {
+            t!("-> RevocationStatus::Revoked({})", revs.len());
             RevocationStatus::Revoked(revs)
         } else if let Some(revs)
-            = check(&self.other_revocations, Default::default())
+            = check(&mut self.other_revocations.iter(), Default::default())
         {
+            t!("-> RevocationStatus::CouldBe({})", revs.len());
             RevocationStatus::CouldBe(revs)
         } else {
+            t!("-> RevocationStatus::NotAsFarAsWeKnow");
             RevocationStatus::NotAsFarAsWeKnow
         }
     }
@@ -721,9 +814,9 @@ impl<C> ComponentBundle<C> {
     {
         let p : Packet = self.component.into();
         std::iter::once(p)
-            .chain(self.self_revocations.into_iter().map(|s| s.into()))
-            .chain(self.self_signatures.into_iter().map(|s| s.into()))
-            .chain(self.attestations.into_iter().map(|s| s.into()))
+            .chain(self.self_revocations.into_unverified().map(|s| s.into()))
+            .chain(self.self_signatures.into_unverified().map(|s| s.into()))
+            .chain(self.attestations.into_unverified().map(|s| s.into()))
             .chain(self.certifications.into_iter().map(|s| s.into()))
             .chain(self.other_revocations.into_iter().map(|s| s.into()))
     }
@@ -770,12 +863,12 @@ impl<C> ComponentBundle<C> {
         // Order self signatures so that the most recent one comes
         // first.
         self.self_signatures.sort_by(sig_cmp);
-        self.self_signatures.iter_mut().for_each(sig_fixup);
+        self.self_signatures.iter_mut_unverified().for_each(sig_fixup);
 
         self.attestations.sort_by(Signature::normalized_cmp);
         self.attestations.dedup_by(sig_merge);
         self.attestations.sort_by(sig_cmp);
-        self.attestations.iter_mut().for_each(sig_fixup);
+        self.attestations.iter_mut_unverified().for_each(sig_fixup);
 
         self.certifications.sort_by(Signature::normalized_cmp);
         self.certifications.dedup_by(sig_merge);
@@ -790,7 +883,7 @@ impl<C> ComponentBundle<C> {
         self.self_revocations.sort_by(Signature::normalized_cmp);
         self.self_revocations.dedup_by(sig_merge);
         self.self_revocations.sort_by(sig_cmp);
-        self.self_revocations.iter_mut().for_each(sig_fixup);
+        self.self_revocations.iter_mut_unverified().for_each(sig_fixup);
 
         self.other_revocations.sort_by(Signature::normalized_cmp);
         self.other_revocations.dedup_by(sig_merge);
@@ -833,7 +926,42 @@ impl<P: key::KeyParts, R: key::KeyRole> ComponentBundle<Key<P, R>> {
     }
 }
 
+impl<P: key::KeyParts> ComponentBundle<Key<P, key::PrimaryRole>> {
+    /// Returns all of the bundles's bad signatures.
+    pub(crate) fn bad_signatures(&self)
+        -> impl Iterator<Item = &Signature> + Send + Sync
+    {
+        self.self_signatures.iter_bad(self.backsig_signer.as_ref())
+            .chain(self.self_revocations.iter_bad(self.backsig_signer.as_ref()))
+    }
+}
+
 impl<P: key::KeyParts> ComponentBundle<Key<P, key::SubordinateRole>> {
+    /// Creates a new subkey component.
+    ///
+    /// Should only be used from the cert parser.  However, we cannot
+    /// use `pub(in ...)` because the cert parser isn't an ancestor of
+    /// this module.
+    pub(crate) fn new_subkey(component: Key<P, key::SubordinateRole>,
+                             hash_algo_security: HashAlgoSecurity,
+                             sigs: Vec<Signature>,
+                             primary_key: Arc<Key<key::PublicParts, key::PrimaryRole>>)
+           -> Self
+    {
+        let backsig_signer = component.pk_algo().for_signing()
+            .then(|| component.parts_as_public().clone());
+        ComponentBundle {
+            component,
+            hash_algo_security,
+            self_signatures: LazySignatures::new(primary_key.clone()),
+            backsig_signer,
+            certifications: sigs,
+            attestations: LazySignatures::new(primary_key.clone()),
+            self_revocations: LazySignatures::new(primary_key),
+            other_revocations: vec![],
+        }
+    }
+
     /// Returns the subkey's revocation status at time `t`.
     ///
     /// A subkey is revoked at time `t` if:
@@ -878,6 +1006,14 @@ impl<P: key::KeyParts> ComponentBundle<Key<P, key::SubordinateRole>> {
         self._revocation_status(policy, t, true,
                                 self.binding_signature(policy, t).ok())
     }
+
+    /// Returns all of the bundles's bad signatures.
+    pub(crate) fn bad_signatures(&self)
+        -> impl Iterator<Item = &Signature> + Send + Sync
+    {
+        self.self_signatures.iter_bad(self.backsig_signer.as_ref())
+            .chain(self.self_revocations.iter_bad(self.backsig_signer.as_ref()))
+    }
 }
 
 impl ComponentBundle<UserID> {
@@ -953,6 +1089,14 @@ impl ComponentBundle<UserID> {
         let t = t.into();
         self._revocation_status(policy, t, false, self.binding_signature(policy, t).ok())
     }
+
+    /// Returns all of the bundles's bad signatures.
+    pub(crate) fn bad_signatures(&self)
+        -> impl Iterator<Item = &Signature> + Send + Sync
+    {
+        self.self_signatures.iter_bad(self.backsig_signer.as_ref())
+            .chain(self.self_revocations.iter_bad(self.backsig_signer.as_ref()))
+    }
 }
 
 impl ComponentBundle<UserAttribute> {
@@ -1024,6 +1168,14 @@ impl ComponentBundle<UserAttribute> {
         self._revocation_status(policy, t, false,
                                 self.binding_signature(policy, t).ok())
     }
+
+    /// Returns all of the bundles's bad signatures.
+    pub(crate) fn bad_signatures(&self)
+        -> impl Iterator<Item = &Signature> + Send + Sync
+    {
+        self.self_signatures.iter_bad(self.backsig_signer.as_ref())
+            .chain(self.self_revocations.iter_bad(self.backsig_signer.as_ref()))
+    }
 }
 
 impl ComponentBundle<Unknown> {
@@ -1053,4 +1205,152 @@ impl ComponentBundle<Unknown> {
     pub fn unknown(&self) -> &Unknown {
         self.component()
     }
+
+    /// Returns all of the bundles's bad signatures.
+    pub(crate) fn bad_signatures(&self)
+        -> impl Iterator<Item = &Signature> + Send + Sync
+    {
+        self.self_signatures.iter_bad(self.backsig_signer.as_ref())
+            .chain(self.self_revocations.iter_bad(self.backsig_signer.as_ref()))
+    }
 }
+
+/// A collection of `ComponentBundles`.
+///
+/// Note: we need this, because we can't `impl Vec<ComponentBundles>`.
+#[derive(Debug, Clone, PartialEq)]
+pub(super) struct ComponentBundles<C>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    bundles: Vec<ComponentBundle<C>>,
+}
+
+impl<C> Default for ComponentBundles<C>
+where
+    ComponentBundle<C>: cmp::PartialEq,
+{
+        fn default() -> Self {
+        ComponentBundles {
+            bundles: vec![],
+        }
+    }
+}
+
+impl<C> Deref for ComponentBundles<C>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    type Target = Vec<ComponentBundle<C>>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.bundles
+    }
+}
+
+impl<C> DerefMut for ComponentBundles<C>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    fn deref_mut(&mut self) -> &mut Vec<ComponentBundle<C>> {
+        &mut self.bundles
+    }
+}
+
+impl<C> From<ComponentBundles<C>> for Vec<ComponentBundle<C>>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    fn from(cb: ComponentBundles<C>) -> Vec<ComponentBundle<C>> {
+        cb.bundles
+    }
+}
+
+impl<C> From<Vec<ComponentBundle<C>>> for ComponentBundles<C>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    fn from(bundles: Vec<ComponentBundle<C>>) -> ComponentBundles<C> {
+        ComponentBundles {
+            bundles,
+        }
+    }
+}
+
+impl<C> IntoIterator for ComponentBundles<C>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    type Item = ComponentBundle<C>;
+    type IntoIter = std::vec::IntoIter<Self::Item>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        self.bundles.into_iter()
+    }
+}
+
+impl<C> ComponentBundles<C>
+    where ComponentBundle<C>: cmp::PartialEq
+{
+    // Sort and dedup the components.
+    //
+    // `cmp` is a function to sort the components for deduping.
+    //
+    // `merge` is a function that merges the first component into the
+    // second component.
+    pub(super) fn sort_and_dedup<F, F2>(&mut self, cmp: F, merge: F2)
+        where F: Fn(&C, &C) -> Ordering,
+              F2: Fn(&mut C, &mut C)
+    {
+        // We dedup by component (not bundles!).  To do this, we need
+        // to sort the bundles by their components.
+
+        self.bundles.sort_by(
+            |a, b| cmp(&a.component, &b.component));
+
+        self.bundles.dedup_by(|a, b| {
+            if cmp(&a.component, &b.component) == Ordering::Equal {
+                // Merge.
+                merge(&mut a.component, &mut b.component);
+
+                // Recall: if a and b are equal, a will be dropped.
+                // Also, the elements are given in the opposite order
+                // from their order in the vector.
+                b.self_signatures.append(&mut a.self_signatures);
+                b.attestations.append(&mut a.attestations);
+                b.certifications.append(&mut a.certifications);
+                b.self_revocations.append(&mut a.self_revocations);
+                b.other_revocations.append(&mut a.other_revocations);
+
+                true
+            } else {
+                false
+            }
+        });
+
+        // And sort the certificates.
+        for b in self.bundles.iter_mut() {
+            b.sort_and_dedup();
+        }
+    }
+}
+
+/// A vecor of key (primary or subkey, public or private) and any
+/// associated signatures.
+pub(super) type KeyBundles<KeyPart, KeyRole>
+    = ComponentBundles<Key<KeyPart, KeyRole>>;
+
+/// A vector of subkeys and any associated signatures.
+pub(super) type SubkeyBundles<KeyPart>
+    = KeyBundles<KeyPart, key::SubordinateRole>;
+
+/// A vector of key (primary or subkey, public or private) and any
+/// associated signatures.
+#[allow(dead_code)]
+pub(super) type GenericKeyBundles
+    = ComponentBundles<Key<key::UnspecifiedParts, key::UnspecifiedRole>>;
+
+/// A vector of User ID bundles and any associated signatures.
+pub(super) type UserIDBundles = ComponentBundles<UserID>;
+
+/// A vector of User Attribute bundles and any associated signatures.
+pub(super) type UserAttributeBundles = ComponentBundles<UserAttribute>;
+
+/// A vector of unknown components and any associated signatures.
+///
+/// Note: all signatures are stored as certifications.
+pub(super) type UnknownBundles = ComponentBundles<Unknown>;
