@@ -3,7 +3,7 @@ use std::fmt;
 use crate::Error;
 use crate::Packet;
 use crate::packet::Tag;
-use crate::cert::bundle::{
+use crate::cert::parser::low_level::bundle::{
     SubkeyBundle,
     UserIDBundle,
     UserAttributeBundle,
