@@ -11,6 +11,16 @@ use num_bigint_dig::{traits::ModInverse, BigUint};
 use rsa::traits::{PrivateKeyParts, PublicKeyParts};
 use rsa::{Pkcs1v15Encrypt, RsaPublicKey, RsaPrivateKey, Pkcs1v15Sign};
 
+use ecdsa::{
+    EncodedPoint,
+    hazmat::{SignPrimitive, VerifyPrimitive},
+};
+use p256::elliptic_curve::{
+    generic_array::GenericArray as GA,
+    ops::Reduce,
+    sec1::FromEncodedPoint,
+};
+
 use crate::{Error, Result};
 use crate::crypto::asymmetric::KeyPair;
 use crate::crypto::backend::interface::Asymmetric;
@@ -57,10 +67,8 @@ impl Asymmetric for super::Backend {
     fn supports_curve(curve: &Curve) -> bool {
         use self::Curve::*;
         match curve {
-            NistP256
+            NistP256 | NistP384 | NistP521
                 => true,
-            NistP384 | NistP521
-                => false,
             Ed25519 | Cv25519
                 => true,
             BrainpoolP256 | BrainpoolP512 | Unknown(_)
@@ -264,18 +272,9 @@ impl KeyPair {
              mpi::SecretKeyMaterial::ECDSA { scalar }) => match curve
             {
                 Curve::NistP256 => {
-                    use p256::{
-                        elliptic_curve::{
-                            generic_array::GenericArray as GA,
-                            ops::Reduce,
-                        },
-                        Scalar,
-                    };
-                    use ecdsa::{
-                        hazmat::SignPrimitive,
-                    };
-
+                    use p256::Scalar;
                     const LEN: usize = 32;
+
                     let key = scalar.value_padded(LEN);
                     let key = Scalar::reduce_bytes(GA::try_from_slice(&key)?);
                     let dig = pad_truncating(digest, LEN);
@@ -296,6 +295,57 @@ impl KeyPair {
                         s: MPI::new(&sig.s().to_bytes()),
                     })
                 },
+
+                Curve::NistP384 => {
+                    use p384::Scalar;
+                    const LEN: usize = 48;
+
+                    let key = scalar.value_padded(LEN);
+                    let key = Scalar::reduce_bytes(GA::try_from_slice(&key)?);
+                    let dig = pad_truncating(digest, LEN);
+                    let dig = GA::try_from_slice(&dig)?;
+
+                    let sig = loop {
+                        let mut k: Protected = vec![0; LEN].into();
+                        crate::crypto::random(&mut k);
+                        let k = Scalar::reduce_bytes(
+                            GA::try_from_slice(&k)?);
+                        if let Ok(s) = key.try_sign_prehashed(k, &dig) {
+                            break s.0;
+                        }
+                    };
+
+                    Ok(mpi::Signature::ECDSA {
+                        r: MPI::new(&sig.r().to_bytes()),
+                        s: MPI::new(&sig.s().to_bytes()),
+                    })
+                },
+
+                Curve::NistP521 => {
+                    use p521::Scalar;
+                    const LEN: usize = 66;
+
+                    let key = scalar.value_padded(LEN);
+                    let key = Scalar::reduce_bytes(GA::try_from_slice(&key)?);
+                    let dig = pad_truncating(digest, LEN);
+                    let dig = GA::try_from_slice(&dig)?;
+
+                    let sig = loop {
+                        let mut k: Protected = vec![0; LEN].into();
+                        crate::crypto::random(&mut k);
+                        let k = Scalar::reduce_bytes(
+                            GA::try_from_slice(&k)?);
+                        if let Ok(s) = key.try_sign_prehashed(k, &dig) {
+                            break s.0;
+                        }
+                    };
+
+                    Ok(mpi::Signature::ECDSA {
+                        r: MPI::new(&sig.r().to_bytes()),
+                        s: MPI::new(&sig.s().to_bytes()),
+                    })
+                },
+
                 _ => Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
             },
 
@@ -430,18 +480,7 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
              mpi::Signature::ECDSA { r, s }) => match curve
             {
                 Curve::NistP256 => {
-                    use p256::{
-                        AffinePoint,
-                        ecdsa::Signature,
-                        elliptic_curve::{
-                            generic_array::GenericArray as GA,
-                            sec1::FromEncodedPoint,
-                        },
-                    };
-                    use ecdsa::{
-                        EncodedPoint,
-                        hazmat::VerifyPrimitive,
-                    };
+                    use p256::{AffinePoint, ecdsa::Signature};
                     const LEN: usize = 32;
 
                     let key = AffinePoint::from_encoded_point(
@@ -463,6 +502,55 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                     let dig = GA::try_from_slice(&dig)?;
                     key.verify_prehashed(&dig, &sig).map_err(bad)
                 },
+
+                Curve::NistP384 => {
+                    use p384::{AffinePoint, ecdsa::Signature};
+                    const LEN: usize = 48;
+
+                    let key = AffinePoint::from_encoded_point(
+                        &EncodedPoint::<p384::NistP384>::from_bytes(q.value())?);
+                    let key = if key.is_some().into() {
+                        key.unwrap()
+                    } else {
+                        return Err(Error::InvalidKey(
+                            "Point is not on the curve".into()).into());
+                    };
+
+                    let sig = Signature::from_scalars(
+                        GA::try_clone_from_slice(
+                            &r.value_padded(LEN).map_err(bad)?)?,
+                        GA::try_clone_from_slice(
+                            &s.value_padded(LEN).map_err(bad)?)?)
+                        .map_err(bad)?;
+                    let dig = pad_truncating(digest, LEN);
+                    let dig = GA::try_from_slice(&dig)?;
+                    key.verify_prehashed(&dig, &sig).map_err(bad)
+                },
+
+                Curve::NistP521 => {
+                    use p521::{AffinePoint, ecdsa::Signature};
+                    const LEN: usize = 66;
+
+                    let key = AffinePoint::from_encoded_point(
+                        &EncodedPoint::<p521::NistP521>::from_bytes(q.value())?);
+                    let key = if key.is_some().into() {
+                        key.unwrap()
+                    } else {
+                        return Err(Error::InvalidKey(
+                            "Point is not on the curve".into()).into());
+                    };
+
+                    let sig = Signature::from_scalars(
+                        GA::try_clone_from_slice(
+                            &r.value_padded(LEN).map_err(bad)?)?,
+                        GA::try_clone_from_slice(
+                            &s.value_padded(LEN).map_err(bad)?)?)
+                        .map_err(bad)?;
+                    let dig = pad_truncating(digest, LEN);
+                    let dig = GA::try_from_slice(&dig)?;
+                    key.verify_prehashed(&dig, &sig).map_err(bad)
+                },
+
                 _ => Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
             },
             _ => Err(Error::MalformedPacket(format!(
@@ -485,7 +573,7 @@ impl<R> Key4<SecretParts, R>
         -> Result<Self> where T: Into<Option<SystemTime>>
     {
         // RFC 4880: `p < q`
-        let (p, q) = if p < q { (p, q) } else { (q, p) };
+        let (p, q) = crate::crypto::rsa_sort_raw_pq(p, q);
 
         // RustCrypto can't compute the public key from the private one, so do it ourselves
         let big_p = BigUint::from_bytes_be(p);
@@ -525,6 +613,8 @@ impl<R> Key4<SecretParts, R>
             [p, q] => (p, q),
             _ => panic!("RSA key generation resulted in wrong number of primes"),
         };
+        // RFC 4880: `p < q`
+        let (p, q) = rsa_sort_pq(p, q);
         let u = p.mod_inverse(q) // RFC 4880: u ≡ p⁻¹ (mod q)
             .and_then(|x| x.to_biguint())
             .expect("rsa crate did not generate coprime p and q");
@@ -607,8 +697,107 @@ impl<R> Key4<SecretParts, R>
                 Ok((PublicKeyAlgorithm::ECDH, public_mpis, private_mpis))
             },
 
+            (Curve::NistP384, true) => {
+                use p384::{EncodedPoint, SecretKey};
+
+                let secret = SecretKey::random(
+                    &mut p384::elliptic_curve::rand_core::OsRng);
+                let public = EncodedPoint::from(secret.public_key());
+
+                let public_mpis = mpi::PublicKey::ECDSA {
+                    curve,
+                    q: MPI::new(public.as_bytes()),
+                };
+                let private_mpis = mpi::SecretKeyMaterial::ECDSA {
+                    scalar: Vec::from(secret.to_bytes().as_slice()).into(),
+                };
+
+                Ok((PublicKeyAlgorithm::ECDSA, public_mpis, private_mpis))
+            },
+
+            (Curve::NistP384, false) => {
+                use p384::{EncodedPoint, SecretKey};
+
+                let secret = SecretKey::random(
+                    &mut p384::elliptic_curve::rand_core::OsRng);
+                let public = EncodedPoint::from(secret.public_key());
+
+                let public_mpis = mpi::PublicKey::ECDH {
+                    q: MPI::new(public.as_bytes()),
+                    hash:
+                    crate::crypto::ecdh::default_ecdh_kdf_hash(&curve),
+                    sym:
+                    crate::crypto::ecdh::default_ecdh_kek_cipher(&curve),
+                    curve,
+                };
+                let private_mpis = mpi::SecretKeyMaterial::ECDH {
+                    scalar: Vec::from(secret.to_bytes().as_slice()).into(),
+                };
+
+                Ok((PublicKeyAlgorithm::ECDH, public_mpis, private_mpis))
+            },
+
+            (Curve::NistP521, true) => {
+                use p521::{EncodedPoint, SecretKey};
+
+                let secret = SecretKey::random(
+                    &mut p521::elliptic_curve::rand_core::OsRng);
+                let public = EncodedPoint::from(secret.public_key());
+
+                let public_mpis = mpi::PublicKey::ECDSA {
+                    curve,
+                    q: MPI::new(public.as_bytes()),
+                };
+                let private_mpis = mpi::SecretKeyMaterial::ECDSA {
+                    scalar: Vec::from(secret.to_bytes().as_slice()).into(),
+                };
+
+                Ok((PublicKeyAlgorithm::ECDSA, public_mpis, private_mpis))
+            },
+
+            (Curve::NistP521, false) => {
+                use p521::{EncodedPoint, SecretKey};
+
+                let secret = SecretKey::random(
+                    &mut p521::elliptic_curve::rand_core::OsRng);
+                let public = EncodedPoint::from(secret.public_key());
+
+                let public_mpis = mpi::PublicKey::ECDH {
+                    q: MPI::new(public.as_bytes()),
+                    hash:
+                    crate::crypto::ecdh::default_ecdh_kdf_hash(&curve),
+                    sym:
+                    crate::crypto::ecdh::default_ecdh_kek_cipher(&curve),
+                    curve,
+                };
+                let private_mpis = mpi::SecretKeyMaterial::ECDH {
+                    scalar: Vec::from(secret.to_bytes().as_slice()).into(),
+                };
+
+                Ok((PublicKeyAlgorithm::ECDH, public_mpis, private_mpis))
+            },
+
             _ => Err(Error::UnsupportedEllipticCurve(curve).into()),
         }
     }
 }
 
+/// Given the secret prime values `p` and `q`, returns the pair of
+/// primes so that the smaller one comes first.
+///
+/// Section 5.5.3 of RFC4880 demands that `p < q`.  This function can
+/// be used to order `p` and `q` accordingly.
+///
+/// Note: even though this function seems trivial, we introduce it as
+/// explicit abstraction.  The reason is that the function's
+/// expression also "works" (as in it compiles) for byte slices, but
+/// does the wrong thing, see [`crate::crypto::rsa_sort_raw_pq`].
+fn rsa_sort_pq<'a>(p: &'a BigUint, q: &'a BigUint)
+                   -> (&'a BigUint, &'a BigUint)
+{
+    if p < q {
+        (p, q)
+    } else {
+        (q, p)
+    }
+}
