@@ -253,19 +253,21 @@ impl KeyHandle {
     /// # use openpgp::KeyID;
     /// # use openpgp::KeyHandle;
     /// #
-    /// # let fpr1 : KeyHandle
+    /// # let fpr1: KeyHandle
     /// #     = "8F17 7771 18A3 3DDA 9BA4  8E62 AACB 3243 6300 52D9"
     /// #       .parse::<Fingerprint>()?.into();
     /// #
-    /// # let fpr2 : KeyHandle
+    /// # let fpr2: KeyHandle
     /// #     = "0123 4567 8901 2345 6789  0123 AACB 3243 6300 52D9"
     /// #       .parse::<Fingerprint>()?.into();
     /// #
-    /// # let keyid : KeyHandle = "AACB 3243 6300 52D9".parse::<KeyID>()?
+    /// # let keyid: KeyHandle = "AACB 3243 6300 52D9".parse::<KeyID>()?
     /// #     .into();
     /// #
     /// // fpr1 and fpr2 are different fingerprints with the same KeyID.
-    /// assert!(! fpr1.eq(&fpr2));
+    /// assert_ne!(fpr1, fpr2);
+    /// assert_eq!(KeyID::from(&fpr1), KeyID::from(&fpr2));
+    ///
     /// assert!(fpr1.aliases(&keyid));
     /// assert!(fpr2.aliases(&keyid));
     /// assert!(! fpr1.aliases(&fpr2));
@@ -274,31 +276,9 @@ impl KeyHandle {
     pub fn aliases<H>(&self, other: H) -> bool
         where H: Borrow<KeyHandle>
     {
-        let other = other.borrow();
-        if self.partial_cmp(other) == Some(Ordering::Equal) {
-            true
-        } else {
-            match (self, other) {
-                (KeyHandle::Fingerprint(Fingerprint::V4(f)),
-                 KeyHandle::KeyID(KeyID::V4(i)))
-                    | (KeyHandle::KeyID(KeyID::V4(i)),
-                       KeyHandle::Fingerprint(Fingerprint::V4(f))) =>
-                {
-                    // Avoid a heap allocation by embedding our
-                    // knowledge of how a v4 key ID is derived from a
-                    // v4 fingerprint:
-                    //
-                    // A v4 key ID are the 8 right-most octets of a v4
-                    // fingerprint.
-                    &f[12..] == i
-                },
-                (KeyHandle::Fingerprint(f), KeyHandle::KeyID(i))
-                    | (KeyHandle::KeyID(i), KeyHandle::Fingerprint(f)) =>
-                {
-                    &KeyID::from(f) == i
-                },
-                _ => false,
-            }
+        match self {
+            KeyHandle::Fingerprint(fpr) => fpr.aliases(other),
+            KeyHandle::KeyID(keyid) => keyid.aliases(other),
         }
     }
 
@@ -341,6 +321,58 @@ impl KeyHandle {
                  | KeyHandle::KeyID(KeyID::Invalid(_)))
     }
 
+    /// Returns whether the KeyHandle contains a fingerprint.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use sequoia_openpgp as openpgp;
+    /// # use openpgp::Fingerprint;
+    /// # use openpgp::KeyID;
+    /// # use openpgp::KeyHandle;
+    /// #
+    /// # fn main() -> sequoia_openpgp::Result<()> {
+    /// let fpr: KeyHandle = "8F17 7771 18A3 3DDA 9BA4  8E62 AACB 3243 6300 52D9"
+    ///     .parse()?;
+    /// let keyid: KeyHandle = KeyHandle::from(KeyID::from(&fpr));
+    ///
+    /// assert!(fpr.is_fingerprint());
+    /// assert!(! keyid.is_fingerprint());
+    /// # Ok(()) }
+    /// ```
+    pub fn is_fingerprint(&self) -> bool {
+        match self {
+            KeyHandle::Fingerprint(_) => true,
+            KeyHandle::KeyID(_) => false,
+        }
+    }
+
+    /// Returns whether the KeyHandle contains a key ID.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use sequoia_openpgp as openpgp;
+    /// # use openpgp::Fingerprint;
+    /// # use openpgp::KeyID;
+    /// # use openpgp::KeyHandle;
+    /// #
+    /// # fn main() -> sequoia_openpgp::Result<()> {
+    /// let fpr: KeyHandle = "8F17 7771 18A3 3DDA 9BA4  8E62 AACB 3243 6300 52D9"
+    ///     .parse()?;
+    /// let keyid: KeyHandle = KeyHandle::from(KeyID::from(&fpr));
+    ///
+    /// assert!(! fpr.is_keyid());
+    /// assert!(keyid.is_keyid());
+    /// # Ok(()) }
+    /// ```
+    pub fn is_keyid(&self) -> bool {
+        match self {
+            KeyHandle::Fingerprint(_) => false,
+            KeyHandle::KeyID(_) => true,
+        }
+    }
+
     /// Converts this `KeyHandle` to its canonical hexadecimal
     /// representation.
     ///
