@@ -138,7 +138,7 @@
 //!         alice.primary_key().component(),
 //!         alice_userid)?;
 //! let alice = alice.insert_packets(alice_certification.clone())?;
-//! # assert!(alice.clone().into_packets().any(|p| {
+//! # assert!(alice.clone().into_packets2().any(|p| {
 //! #   match p {
 //! #       Packet::Signature(sig) => sig == alice_certification,
 //! #       _ => false,
@@ -157,7 +157,7 @@
 //!         bob.primary_key().component(),
 //!         bob_userid)?;
 //! let bob = bob.insert_packets(bob_certification.clone())?;
-//! # assert!(bob.clone().into_packets().any(|p| {
+//! # assert!(bob.clone().into_packets2().any(|p| {
 //! #   match p {
 //! #       Packet::Signature(sig) => sig == bob_certification,
 //! #       _ => false,
@@ -181,7 +181,7 @@
 //!         ca.primary_key().component(),
 //!         ca_userid)?;
 //! let ca = ca.insert_packets(ca_tsig.clone())?;
-//! # assert!(ca.clone().into_packets().any(|p| {
+//! # assert!(ca.clone().into_packets2().any(|p| {
 //! #   match p {
 //! #       Packet::Signature(sig) => sig == ca_tsig,
 //! #       _ => false,
@@ -288,8 +288,8 @@ pub(crate) fn parse_error_downcast(e: ParseError<usize, Token, LexicalError>)
         ParseError::User { error }
         => ParseError::User { error },
 
-        ParseError::UnrecognizedEOF { location, expected }
-        => ParseError::UnrecognizedEOF { location, expected },
+        ParseError::UnrecognizedEof { location, expected }
+        => ParseError::UnrecognizedEof { location, expected },
     }
 }
 
@@ -378,13 +378,32 @@ fn generate_class(caret: bool, chars: impl Iterator<Item=char>) -> Hir
 /// See the [module-level documentation] for more details.
 ///
 ///   [module-level documentation]: self
+///
+/// # A note on equality
+///
+/// We define equality on `Regex` as the equality of the uncompiled
+/// regular expression given to the constructor and whether
+/// sanitizations are enabled.
 #[derive(Clone, Debug)]
 pub struct Regex {
+    /// The original regular expression.
+    ///
+    /// Equality is defined using this and `disable_sanitizations`.
+    re: String,
     regex: regex::Regex,
     disable_sanitizations: bool,
 }
 assert_send_and_sync!(Regex);
 
+impl PartialEq for Regex {
+    fn eq(&self, other: &Self) -> bool {
+        self.re == other.re
+            && self.disable_sanitizations == other.disable_sanitizations
+    }
+}
+
+impl Eq for Regex {}
+
 impl Regex {
     /// Parses and compiles the regular expression.
     ///
@@ -411,6 +430,7 @@ impl Regex {
             .build()?;
 
         Ok(Self {
+            re: re.into(),
             regex,
             disable_sanitizations: false,
         })
@@ -430,6 +450,11 @@ impl Regex {
         Self::new(std::str::from_utf8(re)?)
     }
 
+    /// Returns the string-representation of the regular expression.
+    pub fn as_str(&self) -> &str {
+        &self.re
+    }
+
     /// Controls whether matched strings must pass a sanity check.
     ///
     /// If `false` (the default), i.e., sanity checks are enabled, and
@@ -530,13 +555,32 @@ assert_send_and_sync!(RegexSet_);
 /// See the [module-level documentation] for more details.
 ///
 ///   [module-level documentation]: self
+///
+/// # A note on equality
+///
+/// We define equality on `RegexSet` as the equality of the uncompiled
+/// regular expressions given to the constructor and whether
+/// sanitizations are enabled.
 #[derive(Clone)]
 pub struct RegexSet {
+    /// The original regular expressions.
+    ///
+    /// Equality is defined using this and `disable_sanitizations`.
+    re_bytes: Vec<Vec<u8>>,
     re_set: RegexSet_,
     disable_sanitizations: bool,
 }
 assert_send_and_sync!(RegexSet);
 
+impl PartialEq for RegexSet {
+    fn eq(&self, other: &Self) -> bool {
+        self.re_bytes == other.re_bytes
+            && self.disable_sanitizations == other.disable_sanitizations
+    }
+}
+
+impl Eq for RegexSet {}
+
 impl fmt::Debug for RegexSet {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let mut d = f.debug_struct("RegexSet");
@@ -600,16 +644,15 @@ impl RegexSet {
         let mut had_good = false;
         let mut had_bad = false;
 
+        let mut re_bytes = Vec::new();
         for re in res {
             let re = re.borrow();
+            re_bytes.push(re.as_bytes().into());
+
             let lexer = Lexer::new(re);
             match grammar::RegexParser::new().parse(re, lexer) {
                 Ok(hir) => {
                     had_good = true;
-                    let hir = hir::Hir::group(hir::Group {
-                        kind: hir::GroupKind::NonCapturing,
-                        hir: Box::new(hir),
-                    });
                     regexes.push(hir);
                 }
                 Err(err) => {
@@ -622,6 +665,7 @@ impl RegexSet {
         if had_bad && ! had_good {
             t!("All regular expressions were invalid.");
             Ok(RegexSet {
+                re_bytes,
                 re_set: RegexSet_::Invalid,
                 disable_sanitizations: false,
             })
@@ -629,14 +673,17 @@ impl RegexSet {
             // Match everything.
             t!("No regular expressions provided.");
             Ok(RegexSet {
+                re_bytes,
                 re_set: RegexSet_::Everything,
                 disable_sanitizations: false,
             })
         } else {
             // Match any of the regular expressions.
             Ok(RegexSet {
+                re_bytes,
                 re_set: RegexSet_::Regex(
                     Regex {
+                        re: String::new(),
                         regex: regex::RegexBuilder::new(
                             &Hir::alternation(regexes).to_string())
                             .build()?,
@@ -714,12 +761,14 @@ impl RegexSet {
     {
         let mut have_valid_utf8 = false;
         let mut have_invalid_utf8 = false;
+        let mut re_bytes = Vec::new();
         let re_set = Self::new(
             res
                 .into_iter()
                 .scan((&mut have_valid_utf8, &mut have_invalid_utf8),
                       |(valid, invalid), re|
                       {
+                          re_bytes.push(re.borrow().to_vec());
                           if let Ok(re) = std::str::from_utf8(re.borrow()) {
                               **valid = true;
                               Some(Some(re))
@@ -734,16 +783,22 @@ impl RegexSet {
             // None of the strings were valid UTF-8.  Reject
             // everything.
             Ok(RegexSet {
+                re_bytes,
                 re_set: RegexSet_::Invalid,
                 disable_sanitizations: false,
             })
         } else {
             // We had nothing or at least one string was valid UTF-8.
             // RegexSet::new did the right thing.
-            re_set
+            re_set.map(|mut r| { r.re_bytes = re_bytes; r })
         }
     }
 
+    /// Returns the bytes-representation of the regular expressions.
+    pub fn as_bytes(&self) -> &[Vec<u8>] {
+        &self.re_bytes
+    }
+
     /// Creates a `RegexSet` from the regular expressions stored in a
     /// trust signature.
     ///
@@ -873,6 +928,7 @@ impl RegexSet {
     pub fn everything() -> Result<Self>
     {
         Ok(Self {
+            re_bytes: vec![vec![]],
             re_set: RegexSet_::Everything,
             disable_sanitizations: false,
         })
@@ -1036,7 +1092,6 @@ impl RegexSet {
     /// ```
     pub fn matches_userid(&self, u: &UserID) -> bool
     {
-        let u = u.borrow();
         if let Ok(u) = std::str::from_utf8(u.value()) {
             self.is_match(u)
         } else {
@@ -1411,6 +1466,65 @@ mod tests {
             (true, "xabcdey"),
             (false, "xa(b(c)d)ey"),
         ]);
+        a("x(a|b)y", &[
+            (false, "xy"),
+            (true, "xay"),
+            (true, "xby"),
+            (false, "xaay"),
+            (false, "xbby"),
+            (false, "xaby"),
+            (false, "xaaby"),
+            (false, "xabby"),
+            (false, "xaabby"),
+            (false, "xcy"),
+        ]);
+        a("x(a|bc)y", &[
+            (false, "xy"),
+            (true, "xay"),
+            (false, "xby"),
+            (true, "xbcy"),
+            (false, "xaay"),
+            (false, "xbby"),
+            (false, "xaby"),
+            (false, "xabcy"),
+            (false, "xabby"),
+            (false, "xaabby"),
+            (false, "xcy"),
+            (false, "xacy"),
+        ]);
+        a("x(a|b|c)y", &[
+            (false, "xy"),
+            (true, "xay"),
+            (true, "xby"),
+            (true, "xcy"),
+            (false, "xaay"),
+            (false, "xbby"),
+            (false, "xaby"),
+            (false, "xabcy"),
+            (false, "xabby"),
+            (false, "xaabby"),
+            (false, "xacy"),
+        ]);
+        a("x(a|b)(c|d)y", &[
+            (false, "xy"),
+            (false, "xay"),
+            (false, "xby"),
+            (false, "xcy"),
+            (false, "xdy"),
+            (false, "xaay"),
+            (false, "xbby"),
+            (false, "xccy"),
+            (false, "xddy"),
+            (false, "xaby"),
+            (false, "xcdy"),
+            (true, "xacy"),
+            (true, "xady"),
+            (true, "xbcy"),
+            (true, "xbdy"),
+            (false, "xabcy"),
+            (false, "xabby"),
+            (false, "xaabby"),
+        ]);
         a("x(a+|b+)y", &[
             (false, "xy"),
             (true, "xay"),
@@ -2017,7 +2131,10 @@ mod tests {
         // Try to make sure one re does not leak into another.
         let re = RegexSet::new(&[ "cd$", "^ab" ])?;
         assert!(re.is_match("abxx"));
+        assert!(! re.is_match("xabxx"));
         assert!(re.is_match("xxcd"));
+        assert!(! re.is_match("xxcdx"));
+        assert!(re.is_match("abcdx"));
 
         // Invalid regular expressions should be ignored.
         let re = RegexSet::new(&[ "[ab", "cd]", "x" ])?;
@@ -2052,6 +2169,14 @@ mod tests {
         assert!(re.is_match("cd]"));
         assert!(re.is_match("x"));
 
+        // The empty branch of the alternation should match everything.
+        let re = RegexSet::new(&[ "ab|", "cd" ])?;
+        assert!(re.is_match("a"));
+        assert!(re.is_match("b"));
+        assert!(re.is_match("x"));
+        assert!(re.is_match("xyx"));
+        assert!(re.is_match(""));
+
         Ok(())
     }
 
