@@ -1,6 +1,16 @@
+use std::{
+    borrow::Cow,
+};
+
 use crate::Result;
 use crate::cert::prelude::*;
-use crate::packet::{header::BodyLength, key, Signature, Tag};
+use crate::packet::{
+    header::BodyLength,
+    key,
+    Key,
+    Signature,
+    Tag,
+};
 use crate::seal;
 use crate::serialize::{
     PacketRef,
@@ -11,6 +21,65 @@ use crate::serialize::{
 
 
 impl Cert {
+    /// Returns whether the certificate should be exported.
+    ///
+    /// A certificate should only be exported if it has at least one
+    /// exportable direct key signature, or there is at least one user
+    /// ID with at least one exportable self signature.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use sequoia_openpgp as openpgp;
+    /// use openpgp::cert::prelude::*;
+    ///
+    /// # fn main() -> openpgp::Result<()> {
+    /// // By default, certificates are exportable.
+    /// let (cert, _) =
+    ///     CertBuilder::general_purpose(None, Some("alice@example.org"))
+    ///         .generate()?;
+    /// assert!(cert.exportable());
+    ///
+    /// // Setting the exportable flag to false makes them
+    /// // not-exportable.
+    /// let (cert, _) =
+    ///     CertBuilder::general_purpose(None, Some("alice@example.org"))
+    ///         .set_exportable(false)
+    ///         .generate()?;
+    /// assert!(! cert.exportable());
+    /// # Ok(())
+    /// # }
+    /// ```
+    #[allow(clippy::if_same_then_else)]
+    pub fn exportable(&self) -> bool {
+        let pk = self.primary_key();
+
+        if pk.self_signatures().chain(pk.self_revocations())
+            .any(|sig| sig.exportable().is_ok())
+        {
+            // Exportable direct key signature.  Export it.
+            true
+        } else if self.userids().any(|userid| {
+            userid.self_signatures()
+                .chain(userid.self_revocations())
+                .any(|sig| sig.exportable().is_ok())
+        }) {
+            // User ID with exportable self signature.  Export it.
+            true
+        } else if self.user_attributes().any(|ua| {
+            ua.self_signatures()
+                .chain(ua.self_revocations())
+                .any(|sig| sig.exportable().is_ok())
+        }) {
+            // User attribute with exportable self signature.  Export
+            // it.
+            true
+        } else {
+            // Don't export it.
+            false
+        }
+    }
+
     /// Serializes or exports the Cert.
     ///
     /// If `export` is true, then non-exportable signatures are not
@@ -32,6 +101,10 @@ impl Cert {
     fn serialize_common(&self, o: &mut dyn std::io::Write, export: bool)
                         -> Result<()>
     {
+        if export && ! self.exportable() {
+            return Ok(())
+        }
+
         let primary = self.primary_key();
         PacketRef::PublicKey(primary.key())
             .serialize(o)?;
@@ -201,6 +274,15 @@ impl Cert {
     pub fn as_tsk(&self) -> TSK {
         TSK::new(self)
     }
+
+    /// Derive a [`TSK`] object from this key that owns the `Cert`.
+    ///
+    /// This object writes out secret keys during serialization.
+    ///
+    /// [`TSK`]: crate::serialize::TSK
+    pub fn into_tsk(self) -> TSK<'static> {
+        TSK::from(self)
+    }
 }
 
 /// A reference to a `Cert` that allows serialization of secret keys.
@@ -230,8 +312,8 @@ impl Cert {
 /// # Ok(()) }
 /// ```
 pub struct TSK<'a> {
-    pub(crate) cert: &'a Cert,
-    filter: Box<dyn Fn(&'a key::UnspecifiedSecret) -> bool + 'a>,
+    pub(crate) cert: Cow<'a, Cert>,
+    filter: Box<dyn Fn(&key::UnspecifiedSecret) -> bool + 'a>,
     emit_stubs: bool,
 }
 
@@ -270,16 +352,46 @@ impl PartialEq for TSK<'_> {
     }
 }
 
+impl From<Cert> for TSK<'_> {
+    fn from(c: Cert) -> Self {
+        Self {
+            cert: Cow::Owned(c),
+            filter: Box::new(|_| true),
+            emit_stubs: false,
+        }
+    }
+}
+
 impl<'a> TSK<'a> {
     /// Creates a new view for the given `Cert`.
     fn new(cert: &'a Cert) -> Self {
         Self {
-            cert,
+            cert: Cow::Borrowed(cert),
             filter: Box::new(|_| true),
             emit_stubs: false,
         }
     }
 
+    /// Decomposes the TSK.
+    pub(crate) fn decompose(self)
+                            -> (Cow<'a, Cert>,
+                                Box<dyn 'a + Fn(&key::UnspecifiedSecret) -> bool>,
+                                bool)
+    {
+        (self.cert, self.filter, self.emit_stubs)
+    }
+
+    /// Returns whether we're emitting secret (sub)key packets when
+    /// serializing.
+    ///
+    /// Computes whether we have secrets, taking the filter into
+    /// account, and whether we are emitting stubs.  This can be used
+    /// to determine the correct armor label to use.
+    pub(crate) fn emits_secret_key_packets(&self) -> bool {
+        self.emit_stubs
+            || self.cert.keys().secret().any(|skb| (self.filter)(skb.key()))
+    }
+
     /// Filters which secret keys to export using the given predicate.
     ///
     /// Note that the given filter replaces any existing filter.
@@ -311,7 +423,7 @@ impl<'a> TSK<'a> {
     /// # Ok(()) }
     /// ```
     pub fn set_filter<P>(mut self, predicate: P) -> Self
-        where P: 'a + Fn(&'a key::UnspecifiedSecret) -> bool
+        where P: 'a + Fn(&key::UnspecifiedSecret) -> bool
     {
         self.filter = Box::new(predicate);
         self
@@ -394,6 +506,36 @@ impl<'a> TSK<'a> {
         self
     }
 
+    /// Adds a GnuPG-style secret key stub to the key.
+    pub(crate) fn add_stub<P, R>(key: Key<P, R>) -> key::UnspecifiedSecret
+    where
+        P: key::KeyParts,
+        R: key::KeyRole,
+    {
+        let key = key.parts_into_public().role_into_unspecified();
+
+        // Emit a GnuPG-style secret key stub.
+        let stub = crate::crypto::S2K::Private {
+            tag: 101,
+            parameters: Some(vec![
+                0,    // "hash algo"
+                0x47, // 'G'
+                0x4e, // 'N'
+                0x55, // 'U'
+                1     // "mode"
+            ].into()),
+        };
+        key.add_secret(key::SecretKeyMaterial::Encrypted(
+            key::Encrypted::new(
+                stub, 0.into(),
+                // Mirrors more closely what GnuPG 2.1
+                // does (oddly, GnuPG 1.4 emits 0xfe
+                // here).
+                Some(crate::crypto::mpi::SecretKeyChecksum::Sum16),
+                vec![].into())))
+            .0.into()
+    }
+
     /// Serializes or exports the Cert.
     ///
     /// If `export` is true, then non-exportable signatures are not
@@ -418,10 +560,13 @@ impl<'a> TSK<'a> {
 
         // Serializes public or secret key depending on the filter.
         let serialize_key =
-            |o: &mut dyn std::io::Write, key: &'a key::UnspecifiedSecret,
+            |o: &mut dyn std::io::Write, key: &'a key::UnspecifiedKey,
              tag_public, tag_secret|
         {
-            let tag = if key.has_secret() && (self.filter)(key) {
+            // We check for secrets here.
+            let tag = if key.has_secret()
+                && (self.filter)(key.try_into().expect("checked for secrets"))
+            {
                 tag_secret
             } else {
                 tag_public
@@ -429,26 +574,7 @@ impl<'a> TSK<'a> {
 
             if self.emit_stubs && (tag == Tag::PublicKey
                                    || tag == Tag::PublicSubkey) {
-                // Emit a GnuPG-style secret key stub.
-                let stub = crate::crypto::S2K::Private {
-                    tag: 101,
-                    parameters: Some(vec![
-                        0,    // "hash algo"
-                        0x47, // 'G'
-                        0x4e, // 'N'
-                        0x55, // 'U'
-                        1     // "mode"
-                    ].into()),
-                };
-                let key_with_stub = key.clone()
-                    .add_secret(key::SecretKeyMaterial::Encrypted(
-                        key::Encrypted::new(
-                            stub, 0.into(),
-                            // Mirrors more closely what GnuPG 2.1
-                            // does (oddly, GnuPG 1.4 emits 0xfe
-                            // here).
-                            Some(crate::crypto::mpi::SecretKeyChecksum::Sum16),
-                            vec![].into()))).0;
+                let key_with_stub = Self::add_stub(key.clone());
                 return match tag {
                     Tag::PublicKey =>
                         crate::Packet::SecretKey(key_with_stub.into())
@@ -465,14 +591,24 @@ impl<'a> TSK<'a> {
                     PacketRef::PublicKey(key.into()).serialize(o),
                 Tag::PublicSubkey =>
                     PacketRef::PublicSubkey(key.into()).serialize(o),
-                Tag::SecretKey =>
-                    PacketRef::SecretKey(key.into()).serialize(o),
-                Tag::SecretSubkey =>
-                    PacketRef::SecretSubkey(key.into()).serialize(o),
+                Tag::SecretKey => {
+                    PacketRef::SecretKey(
+                        key.try_into().expect("checked for secrets"))
+                        .serialize(o)
+                }
+                Tag::SecretSubkey => {
+                    PacketRef::SecretSubkey(
+                        key.try_into().expect("checked for secrets"))
+                    .serialize(o)
+                }
                 _ => unreachable!(),
             }
         };
 
+        if export && ! self.cert.exportable() {
+            return Ok(())
+        }
+
         let primary = self.cert.primary_key();
         serialize_key(o, primary.key().into(),
                       Tag::PublicKey, Tag::SecretKey)?;
@@ -568,9 +704,10 @@ impl<'a> MarshalInto for TSK<'a> {
 
         // Serializes public or secret key depending on the filter.
         let serialized_len_key
-            = |key: &'a key::UnspecifiedSecret, tag_public, tag_secret|
+            = |key: &'a key::UnspecifiedKey, tag_public, tag_secret|
         {
-            let tag = if key.has_secret() && (self.filter)(key) {
+            // We check for secrets here.
+            let tag = if key.has_secret() && (self.filter)(key.try_into().expect("have secrets")) {
                 tag_secret
             } else {
                 tag_public
@@ -589,8 +726,14 @@ impl<'a> MarshalInto for TSK<'a> {
             let packet = match tag {
                 Tag::PublicKey => PacketRef::PublicKey(key.into()),
                 Tag::PublicSubkey => PacketRef::PublicSubkey(key.into()),
-                Tag::SecretKey => PacketRef::SecretKey(key.into()),
-                Tag::SecretSubkey => PacketRef::SecretSubkey(key.into()),
+                Tag::SecretKey => {
+                    PacketRef::SecretKey(
+                        key.try_into().expect("checked for secrets"))
+                }
+                Tag::SecretSubkey => {
+                    PacketRef::SecretSubkey(
+                        key.try_into().expect("checked for secrets"))
+                }
                 _ => unreachable!(),
             };
 
@@ -919,9 +1062,11 @@ mod test {
             ($a: expr, $b: expr, $expectation: expr) => {{
                 let a = $a;
                 let b = $b;
-                let serialized_eq = a.to_vec()? == b.to_vec()?;
                 let eq = a == b;
+                let serialized_eq = a.to_vec()? == b.to_vec()?;
+                let armored_eq = a.armored().to_vec()? == b.armored().to_vec()?;
                 assert_eq!(serialized_eq, eq);
+                assert_eq!(armored_eq, eq);
                 assert_eq!(eq, $expectation);
             }};
         }
@@ -933,7 +1078,7 @@ mod test {
         check!(&cert_0, &tsk_0, true);
 
         // Filters out secrets.
-        let no_secrets = |_| false;
+        let no_secrets = |_: &_| false;
 
         // TSK's equality.
         check!(tsk_0.as_tsk(), tsk_1.as_tsk(), true);
@@ -967,4 +1112,60 @@ mod test {
 
         Ok(())
     }
+
+    #[test]
+    fn issue_1075() -> Result<()> {
+        let cert = Cert::from_bytes(crate::tests::key("testy.pgp"))?;
+        let tsk = Cert::from_bytes(crate::tests::key("testy-private.pgp"))?;
+
+        // Filters out secrets.
+        let no_secrets = |_: &_| false;
+
+        assert!(! cert.as_tsk().emits_secret_key_packets());
+        assert!(cert.as_tsk().emit_secret_key_stubs(true)
+                .emits_secret_key_packets());
+        assert!(tsk.as_tsk().emits_secret_key_packets());
+        assert!(! tsk.as_tsk().set_filter(no_secrets)
+                .emits_secret_key_packets());
+
+        assert!(cert.armored().to_vec()? != tsk.as_tsk().armored().to_vec()?);
+        assert_eq!(cert.armored().to_vec()?,
+                   cert.as_tsk().armored().to_vec()?);
+        assert_eq!(cert.armored().to_vec()?,
+                   tsk.as_tsk().set_filter(no_secrets).armored().to_vec()?);
+
+        Ok(())
+    }
+
+    #[test]
+    fn into_packets() -> Result<()> {
+        let cert = Cert::from_bytes(crate::tests::key("testy-private.pgp"))?;
+
+        fn t(tsk: TSK) {
+            let a = tsk.to_vec().unwrap();
+            let b = {
+                let mut b = Vec::new();
+                tsk.into_packets().for_each(|p| p.serialize(&mut b).unwrap());
+                b
+            };
+            assert_eq!(a, b);
+        }
+
+        let tsk = cert.clone().into_tsk();
+        t(tsk);
+
+        let tsk = cert.clone().into_tsk().set_filter(|_| false);
+        t(tsk);
+
+        let tsk = cert.clone().into_tsk()
+            .set_filter(|k| k.fingerprint() == cert.fingerprint());
+        t(tsk);
+
+        let tsk = cert.clone().into_tsk()
+            .set_filter(|k| k.fingerprint() == cert.fingerprint())
+            .emit_secret_key_stubs(true);
+        t(tsk);
+
+        Ok(())
+    }
 }
